#  Copyright (c) 2022.  NetFoundry, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import socket
from socket import getaddrinfo as PyGetaddrinfo
from socket import socket as PySocket
from typing import Tuple, Union

from . import context, zitilib


class ZitiSocket(PySocket):
    # pylint: disable=redefined-builtin
    def __init__(self, af=-1, type=-1, proto=-1, fileno=None, opts=None):
        zitilib.init()
        if opts is None:
            opts = {}
        if opts.get('bindings') is None:
            opts['bindings'] = {}
        self._bind_address = None
        self._ziti_opts = opts
        self._ziti_af = af
        self._ziti_type = type
        self._ziti_proto = proto
        if fileno:
            super().__init__(af, type, proto, fileno)
            return

        if type == -1:
            type = socket.SOCK_STREAM

        self._zitifd = zitilib.ziti_socket(type)
        super().__init__(af, type, proto, self._zitifd)

    def connect(self, addr) -> None:
        if self._zitifd is None:
            pass
        if isinstance(addr, tuple):
            try:
                zitilib.connect_addr(self._zitifd, addr)
            except:
                PySocket.close(self)
                self._zitifd = None
                PySocket.__init__(self, self._ziti_af, self._ziti_type, self._ziti_proto)
                PySocket.connect(self, addr)

    def close(self) -> None:
        try:
            zitifd = getattr(self, '_zitifd')
        except AttributeError:
            super().close()
        else:
            zitilib.ziti_close(zitifd)

    def bind(self, addr) -> None:
        self._bind_address = addr
        bindings = self._ziti_opts['bindings']
        cfg = bindings.get(addr)
        if cfg is not None:
            ztx = context.get_context(cfg['ztx'])
            service = cfg['service']
            terminator = None
            if isinstance(service, tuple):
                service, terminator = service
            ztx.bind(service=service, terminator=terminator, sock=self)
        else:
            PySocket.close(self)
            self._zitifd = None
            PySocket.__init__(self, self._ziti_af, self._ziti_type, self._ziti_proto)
            PySocket.bind(self, addr)

    def getsockname(self) -> Tuple['str', int]:
        # return this for now since frameworks expect something to be returned
        return ('127.0.0.1', 0)

    def listen(self, __backlog: int = 5) -> None:
        try:
            zitilib.listen(self._zitifd, __backlog)
        except:
            super().listen(__backlog)

    def accept(self):
        fd, peer = zitilib.accept(self.fileno())
        return ZitiSocket(af=self._ziti_af, type=self._ziti_type, fileno=fd), peer

    def setsockopt(self, __level: int, __optname: int, __value: Union[int, bytes]) -> None:
        try:
            PySocket.setsockopt(self, __level, __optname, __value)
        except:
            pass


def create_ziti_connection(address, **_):
    sock = ZitiSocket(socket.SOCK_STREAM)
    sock.connect(address)
    return sock


def ziti_getaddrinfo(host, port, family=0, type=0, proto=0, flags=0):
    # pylint: disable=redefined-builtin
    addrs = zitilib.getaddrinfo(host, port, family, type, proto, flags)
    if addrs is None:
        addrs = PyGetaddrinfo(host, port, family, type, proto, flags)
    return addrs



#  Copyright (c)  NetFoundry Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from os import getenv

from . import _version, context, decor, zitilib, zitisock

_ziti_identities = filter(lambda p: p != '',
                          map(lambda s: s.strip(),
                              (getenv('ZITI_IDENTITIES') or "").split(';')))

enroll = zitilib.enroll
version = zitilib.version
shutdown = zitilib.shutdown
load = context.load_identity
socket = zitisock.ZitiSocket

for identity in _ziti_identities:
    if identity != '':
        load(identity)

monkeypatch = decor.MonkeyPatch
zitify = decor.zitify

__version__ = _version.get_versions()['version']
del _version



#  Copyright (c) 2022.  NetFoundry, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import socket
from os.path import isdir, isfile

from . import zitilib, zitisock


class ZitiContext:
    """
    Object representing a Ziti Identity.
    """
    def __init__(self, ctx):
        if not isinstance(ctx, int):
            raise TypeError("ctx is not a valid python void pointer type")
        self._ctx = ctx

    def connect(self, addr, terminator=None):
        """
        Connect to a Ziti service.
        Either service name or intercept address(host,port)
        could be used to connect to service.

        :param addr: service name (string) or
            service intercept address(tuple[host, port])
        :param terminator: specific terminator for the given service,
            ignored if [addr] is an intercept address
        :return: socket connected to the service
        """
        fd = zitilib.ziti_socket(socket.SOCK_STREAM)
        if isinstance(addr, str):
            zitilib.connect(fd, self._ctx, service=addr, terminator=terminator)
        elif isinstance(addr, tuple):
            zitilib.connect_addr(fd, addr)
        else:
            raise TypeError(f'unsupported address {addr}')
        return socket.socket(socket.AF_UNIX, socket.SOCK_STREAM, 0, fd)

    def bind(self, service, terminator=None, sock=None):
        """
        Bind to a Ziti service and return server socket.

        :param service: service name to bind to
        :param terminator: (optional) bind with specific terminator
        :param sock: existing socket to bind
        :return: server socket ready for accept() call
        """
        if sock is None:
            sock = zitisock.ZitiSocket(type=socket.SOCK_STREAM)
        zitilib.bind(sock.fileno(), self._ctx, service, terminator)
        return sock

    @classmethod
    def from_path(cls, path):
        """
        Load Ziti Identity

        :param path: path to Ziti Identity file
        :return: ZitiContext representing given identity
        """
        if not isinstance(path, str):
            raise TypeError("path must be a string")
        if not (isfile(path) or isdir(path)):
            raise ValueError(f"{path} is not a valid path")
        return cls(zitilib.load(path))


def load_identity(path) -> ZitiContext:
    """
    Load Ziti Identity

    :param path: path to Ziti Identity file
    :return: Ziti Context object representing Ziti Identity
    """
    return ZitiContext.from_path(path)


def get_context(ztx) -> ZitiContext:
    if isinstance(ztx, ZitiContext):
        return ztx
    if isinstance(ztx, str):
        return ZitiContext.from_path(ztx)
    raise TypeError(f'{ztx} is not a ZitiContext or str instance')



#  Copyright (c)  NetFoundry Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import socket as sock

from . import zitisock

_patch_methods = {
    "create_connection": zitisock.create_ziti_connection,
    "getaddrinfo": zitisock.ziti_getaddrinfo
}


def _patchedSocket(patch_opts):
    class patchedSocket(zitisock.ZitiSocket):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **dict(kwargs, opts=patch_opts))
    return patchedSocket


class MonkeyPatch:
    """
    Monkeypatch allow to connect to or host Ziti services
    when application does not manage its own sockets.

    Most application won't open sockets directly but will use a library
    or a framework to do the low level networking work for it.

    `openziti.monkeypatch()` makes it possible to use Ziti sockets
    when you don't have access or want to modify the code
    opening sockets or connections.

        Typical usage:

        import openziti, requests
        with openziti.monkeypatch():
            r = requests.get('http://httpbin.ziti/json')
            print(r.headers)
            print(r.json())
    """
    def __init__(self, **kwargs):
        self.orig_socket = sock.socket
        sock.socket = _patchedSocket(kwargs)
        self.orig_methods = {m: sock.__dict__[m] for m, _ in
                             _patch_methods.items()}
        for m_name, _ in _patch_methods.items():
            sock.__dict__[m_name] = _patch_methods[m_name]

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for m_name, _ in self.orig_methods.items():
            sock.__dict__[m_name] = self.orig_methods[m_name]


def zitify(**zkwargs):
    """Decorate a function to monkeypatch its execution with openziti.

    Sample usage:

    @openziti.zitify()
    def do_get(url):
        r = requests.get(url)
        return r.json()
    """
    def zitify_func(func):
        def zitified(*args, **kwargs):
            with MonkeyPatch(**zkwargs):
                func(*args, **kwargs)
        return zitified
    return zitify_func



#  Copyright (c) 2022.  NetFoundry, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# pylint: disable=no-member

import ctypes
import os
import platform
import socket
from typing import Optional, Tuple

_mod_path = os.path.dirname(__file__)
osname = platform.system().lower()

if osname == 'linux':
    LIBNAME = 'libziti.so'
elif osname == 'darwin':
    LIBNAME = 'libziti.dylib'
elif osname == 'windows':
    LIBNAME = 'ziti.dll'
else:
    raise ImportError("could not load ziti shared library")

zitilib_path = _mod_path + f'/lib/{LIBNAME}'
ziti = ctypes.CDLL(zitilib_path)


class _Ver(ctypes.Structure):
    _fields_ = [('version', ctypes.c_char_p), ('revision', ctypes.c_char_p)]

    def __repr__(self):
        return f'({self.version}, {self.revision})'


class SockAddrIn(ctypes.Structure):
    """
    maps struct sockaddr_in

    NOTE:
    -----
    On Linux/Win32 the first two bytes are address family as short,
    on Darwin the first two bytes.
    """
    _fields_ = [
        ('_family', ctypes.c_uint8 * 2),
        ('_port', ctypes.c_uint16),
        ('_addr', ctypes.c_uint8 * 4)
    ]

    def ip(self):
        return '.'.join(str(o) for o in self._addr)

    def af(self):
        fam = self._family[0]
        if osname == 'darwin':
            fam = self._family[1]
        return socket.AddressFamily(fam)

    @property
    def port(self):
        return socket.ntohs(self._port)
    def __repr__(self):
        return f'{self.af().name}:{self.ip()}:{self.port}'


class _AddrInfo(ctypes.Structure):
    """
    int ai_flags;              /* Input flags. */
    int ai_family;             /* Protocol family for socket. */
    int ai_socktype;           /* Socket type. */
    int ai_protocol;           /* Protocol for socket. */
    socklen_t ai_addrlen;      /* Length of socket address. */
    struct sockaddr *ai_addr;  /* Socket address for socket. */
    char *ai_canonname;        /* Canonical name for service location. */
    struct addrinfo *ai_next;  /* Pointer to next in list. */

    NOTE:
    -----
    The order of ai_addr and ai_canonname is switched on
    Darwin and Windows compared to Linux.
    """
    _fields_ = [
        ('ai_flags', ctypes.c_int),
        ('ai_family', ctypes.c_int),
        ('ai_socktype', ctypes.c_int),
        ('ai_protocol', ctypes.c_int),
        ('ai_addrlen', ctypes.c_int32),
        ('ai_p1', ctypes.c_void_p),
        ('ai_p2', ctypes.c_void_p),
        ('ai_next', ctypes.c_void_p)
    ]

    def get_addr(self):
        addr_p = self.ai_p2
        if osname == 'linux':
            addr_p = self.ai_p1
        if addr_p is None:
            return None
        return SockAddrIn.from_address(addr_p)

    def get_canonname(self):
        p = self.ai_p1
        if osname == 'linux':
            p = self.ai_p2
        if p is None:
            return None
        return str(ctypes.cast(p, ctypes.c_char_p))

    def __repr__(self):
        return '|'.join((x[0] + '=' + str(getattr(self, x[0])))
                        for x in _AddrInfo._fields_)


_ziti_version = ziti.ziti_get_version
_ziti_version.restype = ctypes.POINTER(_Ver)

_ziti_lasterr = ziti.Ziti_last_error

_ziti_errorstr = ziti.ziti_errorstr
_ziti_errorstr.argtypes = [ctypes.c_int]
_ziti_errorstr.restype = ctypes.c_char_p

_load_ctx = ziti.Ziti_load_context
_load_ctx.argtypes = [ctypes.POINTER(ctypes.c_char), ]
_load_ctx.restype = ctypes.c_void_p

_ziti_socket = ziti.Ziti_socket
_ziti_socket.argtypes = [ctypes.c_int]
_ziti_socket.restype = ctypes.c_int


_ziti_close = ziti.Ziti_close
_ziti_close.restype = None
_ziti_close.argtypes = [ctypes.c_int]


_ziti_connect = ziti.Ziti_connect
_ziti_connect.argtypes = [ctypes.c_int,     # socket fd
                          ctypes.c_void_p,  # zitt context
                          ctypes.c_char_p,  # service
                          ctypes.c_char_p   # terminator
                          ]

_ziti_connect_addr = ziti.Ziti_connect_addr
_ziti_connect_addr.argtypes = [ctypes.c_int, ctypes.c_char_p, ctypes.c_int]
_ziti_connect_addr.restype = ctypes.c_int

_ziti_bind = ziti.Ziti_bind
_ziti_bind.argtypes = [ctypes.c_int,     # socket fd
                       ctypes.c_void_p,  # ziti context
                       ctypes.c_char_p,  # service
                       ctypes.c_char_p   # terminator
                       ]

_ziti_listen = ziti.Ziti_listen
_ziti_listen.argtypes = [ctypes.c_int, ctypes.c_int]

_ziti_accept = ziti.Ziti_accept
_ziti_accept.argtypes = [ctypes.c_int, ctypes.c_char_p, ctypes.c_int]

_ziti_enroll = ziti.Ziti_enroll_identity
_ziti_enroll.argtypes = [
        ctypes.c_char_p,
        ctypes.c_char_p,
        ctypes.c_char_p,
        ctypes.POINTER(ctypes.c_char_p),
        ctypes.POINTER(ctypes.c_size_t)
]
_ziti_enroll.restype = ctypes.c_int

_ziti_resolve = ziti.Ziti_resolve
_ziti_resolve.argtypes = [
    ctypes.c_char_p,
    ctypes.c_char_p,
    ctypes.POINTER(_AddrInfo),
    ctypes.c_void_p
]


def free_win32(_):
    pass


if osname != 'windows':
    _free = ziti.free
    _free.argtypes = [ctypes.c_void_p]
else:
    _free = free_win32


def version():
    ver = _ziti_version().contents
    return ver


def errorstr(code):
    msg = _ziti_errorstr(code)
    return msg.decode()


def check_error(code):
    if code != 0:
        err = _ziti_lasterr()
        if err < 0:
            msg = _ziti_errorstr(err).decode(encoding='utf-8')
        else:
            msg = errorstr(err)
        raise Exception(err, msg)


def init():
    ziti.Ziti_lib_init()


def ziti_socket(type):
    fd = _ziti_socket(type)
    return fd


def ziti_close(fd):
    _ziti_close(fd)


def shutdown():
    ziti.Ziti_lib_shutdown()


def load(path):
    init()
    b_obj = bytes(path, encoding="utf-8")
    return _load_ctx(b_obj)


def connect(fd, ztx, service: str, terminator: Optional[str] = None):
    srv = bytes(service, encoding='utf-8')
    terminator_b = None
    if terminator:
        terminator_b = bytes(terminator, encoding='utf-8')
    check_error(_ziti_connect(fd, ztx, srv, terminator_b))


def connect_addr(fd, addr: Tuple[str, int]):
    host = bytes(addr[0], encoding='utf-8')
    port = addr[1]
    check_error(_ziti_connect_addr(fd, host, port))


def bind(fd, ztx, service: str, terminator: Optional[str] = None):
    srv = bytes(service, encoding='utf-8')
    terminator_b = None
    if terminator:
        terminator_b = bytes(terminator, encoding='utf-8')
    check_error(_ziti_bind(fd, ztx, srv, terminator_b))


def listen(fd, backlog):
    check_error(_ziti_listen(fd, backlog))


def accept(fd):
    b = ctypes.create_string_buffer(128)
    clt = _ziti_accept(fd, b, 128)
    if clt < 0:
        check_error(clt)
    return clt, (bytes(b.value).decode('utf-8'), 0)


def enroll(jwt, key=None, cert=None):
    """
    Enroll Ziti Identity
    :param jwt: (required) enrollment token,
        can be either name of the token file or a string containing JWT
    :param key: private key to use for enrollment
        (required for 3rd party CA enrollment, otherwise optional,
        new key is generated if None)
    :param cert: certificate to use for enrollment
        (required for 3rd party CA enrollment)
    :return: string containing Ziti Identity in JSON format
    """
    init()
    try:
        with open(jwt, 'rb') as jwt_f:
            jwtc = bytes(jwt_f.read())
    except:
        jwtc = bytes(jwt, 'utf-8')

    id_json = ctypes.c_char_p()
    id_json_len = ctypes.c_size_t()
    retcode = _ziti_enroll(jwtc, key, cert,
                           ctypes.byref(id_json),
                           ctypes.byref(id_json_len))
    if retcode != 0:
        raise RuntimeError(errorstr(retcode))
    try:
        return id_json.value.decode()
    finally:
        _free(id_json)


def getaddrinfo(host, port, family=0, type=0, proto=0, flags=0):
    # pylint: disable=redefined-builtin
    if not isinstance(host, bytes):
        host = bytes(str(host), 'utf-8')
    if not isinstance(port, bytes):
        port = bytes(str(port), 'utf-8')

    hints = _AddrInfo(ai_family=family, ai_socktype=type, ai_protocol=proto, ai_flags=flags)
    addr_p = ctypes.c_void_p()
    rc = _ziti_resolve(host, port, hints, ctypes.byref(addr_p))

    if rc != 0:
        return None

    addr_p = addr_p.value
    result = []
    while addr_p:
        addr = _AddrInfo.from_address(addr_p)
        addr_in = addr.get_addr()
        try:
            af = socket.AddressFamily(addr.ai_family)
        except:
            af = addr.ai_family
        try:
            t = socket.SocketKind(addr.ai_socktype)
        except:
            t = addr.ai_socktype

        a = (af, t, addr.ai_protocol, addr.get_canonname(), (addr_in.ip(), addr_in.port))
        result.append(a)
        addr_p = addr.ai_next

    return result



#  Copyright (c)  NetFoundry Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from . import zitilib


def do_enroll():
    with open(args.jwt) as jwt_file:  # pylint: disable=unspecified-encoding
        jwt = jwt_file.read()

    with open(args.identity, 'wb') as id_file:
        id_json = zitilib.enroll(jwt, key=args.key, cert=args.cert)
        id_file.write(bytes(id_json, 'utf-8'))


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser('openziti')
    parser.set_defaults(func=parser.print_help)
    subcommands = parser.add_subparsers(help='sub-command help',
                                        dest='subcommand')

    enroll_cmd = subcommands.add_parser('enroll', help='enroll identity')
    enroll_cmd.add_argument('-j', '--jwt', required=True)
    enroll_cmd.add_argument('-i', '--identity', required=True)
    enroll_cmd.add_argument('-k', '--key')
    enroll_cmd.add_argument('-c', '--cert')
    enroll_cmd.set_defaults(func=do_enroll)

    args = parser.parse_args()
    if args.subcommand is None:
        parser.print_help()
    else:
        args.func()



#  Copyright (c) 2022.  NetFoundry, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import platform
from setuptools import setup
from setuptools.command.build_ext import build_ext
from setuptools.extension import Extension
from urllib.request import Request, urlopen
import zipfile
from io import BytesIO

import versioneer

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

ZITI_SDK_BASE = 'https://github.com/openziti/ziti-sdk-c/releases/download'


class GetZitilib(build_ext):

    def build_extension(self, ext) -> None:
        ziti_ver = self.get_sdk_version()
        osname, arch, libname = self.get_platform()
        sdk_distro = self.download_sdk(ziti_ver, osname, arch)
        self.extract_zitilib(sdk_distro, libname, self.build_lib)

    def get_platform(self):
        osname = platform.system()
        mach = platform.machine()
        arch, _ = platform.architecture()

        if osname == 'Linux':
            if mach.startswith('arm'):
                if arch == '32bit':
                    mach = 'arm'
                elif arch == '64bit':
                    mach = 'arm64'
            elif mach == 'aarch64':
                mach = 'arm64'
            return osname, mach, 'libziti.so'

        if osname == 'Darwin':
            return osname, mach, 'libziti.dylib'

        if osname == 'Windows':
            return osname, mach, 'ziti.dll'

    def get_sdk_version(self):
        opts = self.distribution.get_option_dict('openziti')
        _, ver = opts['ziti_sdk_version']
        return ver

    def extract_zitilib(self, distro, libname, target):
        with zipfile.ZipFile(BytesIO(distro)) as zipf:
            return zipf.extract(member=f'lib/{libname}', path=f'{target}/openziti')

    def download_sdk(self, version, osname, arch):
        filename = f'{ZITI_SDK_BASE}/{version}/ziti-sdk-{version}-{osname}-{arch}.zip'
        headers = {}
        req = Request(url=filename, headers=headers)
        with urlopen(req) as response:
            length = response.getheader('content-length')
            if response.status != 200:
                raise Exception(f'Could not download "{filename}"')
            print(f"Downloading {length} from {filename}")
            return response.read()

class ZitilibExt(Extension):
    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])


cmds = dict(build_ext=GetZitilib)
cmds = versioneer.get_cmdclass(cmds)

setup(
    version=versioneer.get_version(),
    cmdclass=cmds,
    ext_modules=[
        ZitilibExt('_get_ziti_lib'),
    ],

    packages=['openziti'],
)